
//
// C++ Implementation: decoder
//
// Description:
//
//
// Author: nlm <nlm@hsc.ge>, (C) 2007
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "../Includes/qstdinc.h"
#include "../Includes/qsock.h"
#include "../Includes/qlogger.h"

QLOG_REGISTER_SOURCE

QTCPSock::QTCPSock(QNetAddr* pAddr)
{
	pAddr->CopyTo(&m_bind);

	m_nError = 0;
#ifndef _WIN32
	m_sock = -1;
#else
	m_sock = INVALID_SOCKET;
#endif
	m_bTimed = false;
	m_nTimeOut = 60000;
	m_nSendChunk = 32 * 1024;
	m_nRecvChunk = 32 * 1024;
}

#ifndef _WIN32
QTCPSock::QTCPSock(int sock)
#else
QTCPSock::QTCPSock(SOCKET sock)
#endif
{
	m_nError = 0;
	m_sock = sock;
	m_bTimed = false;
	m_nTimeOut = 60000;
	m_nSendChunk = 32 * 1024;
	m_nRecvChunk = 32 * 1024;
	m_bind.SetAddr(sock);
}

QTCPSock::~QTCPSock()
{
	Close();
}

bool QTCPSock::IsOpen()
{
#ifndef _WIN32
	return m_sock >= 0;
#else
	return m_sock != INVALID_SOCKET;
#endif
}

void QTCPSock::Close()
{
#ifndef _WIN32
	if (m_sock >= 0)
	{
		shutdown(m_sock, SHUT_RDWR);
		close(m_sock);
		m_sock = -1;
	}
#else
	if (m_sock != INVALID_SOCKET)
	{
		shutdown(m_sock, SD_BOTH);
		closesocket(m_sock);
		m_sock = INVALID_SOCKET;
	}
#endif
}

bool QTCPSock::SendByte(uint8_t nData)
{
	return GenSend(&nData, sizeof(nData));
}

bool QTCPSock::SendWord(uint16_t nData)
{
	return GenSend(&nData, sizeof(nData));
}

bool QTCPSock::SendDWord(uint32_t nData)
{
	return GenSend(&nData, sizeof(nData));
}

bool QTCPSock::SendQWord(uint64_t nData)
{
	return GenSend(&nData, sizeof(nData));
}

bool QTCPSock::SendString(const char* pData)
{
	uint16_t nLen = (uint16_t)(strlen(pData) + 1);
	if (!SendWord(nLen)) return false;
	return GenSend(( void*) pData, nLen);
}

bool QTCPSock::SendBool(bool bData)
{
	uint16_t nData = bData?(uint8_t)1:(uint8_t)0;
	return SendWord(nData);
}

bool QTCPSock::SendFloat(float fData)
{
	char buff[32];
	sprintf(buff, "%f", fData);
	return SendString(buff);
}

bool QTCPSock::SendStr(std::string& str)
{
	uint16_t nSize = (uint16_t)str.length();
	if (!SendWord(nSize)) return false;
	if (nSize == 0) return true;
	return GenSend((void*)str.data(), nSize);
}

bool QTCPSock::RecvByte(uint8_t& nData)
{
	return GenRecv(&nData, sizeof(nData));
}

bool QTCPSock::RecvWord(uint16_t& nData)
{
	return GenRecv(&nData, sizeof(nData));
}

bool QTCPSock::RecvDWord(uint32_t& nData)
{
	return GenRecv(&nData, sizeof(nData));
}

bool QTCPSock::RecvQWord(uint64_t& nData)
{
	return GenRecv(&nData, sizeof(nData));
}

bool QTCPSock::RecvString(char* pData)
{
	uint16_t nLen = 0;
	if (!RecvWord(nLen)) return false;
	return GenRecv(pData, nLen);
}

bool QTCPSock::RecvBool(bool& bData)
{
	uint16_t nData = bData?(uint16_t)1:(uint16_t)0;
	if (!RecvWord(nData)) return false;
	bData = nData?true:false;
	return true;
}

bool QTCPSock::RecvFloat(float& fData)
{
	char buff[32];
	float fTmp;
	if (!RecvString(buff))
	{
		return false;
	}
	sscanf(buff, "%f", &fTmp);
	fData = fTmp;
	return true;
}

bool QTCPSock::RecvStr(std::string& str)
{
	uint16_t nSize;
	if (!RecvWord(nSize)) return false;
	if (nSize == 0)
	{
		str.clear();
		return true;		
	}
	str.resize(nSize, 0);
	if (!GenRecv((void*)str.data(), nSize)) return false;
	str = str.c_str();
	return true;
}

bool QTCPSock::GenRecv(void* pData, int nLen)
{
	if (m_bTimed)
	{
		return GenRecvTimed(pData, nLen, m_nTimeOut);
	}

	int nDone = 0;
	char* buf =(char*) pData;
	if (nLen == 0) return true;

	while(nDone < nLen)
	{
		int nChunk = 
#ifndef _WIN32
			std::min(nLen - nDone, m_nRecvChunk);
		int nRecv = recv(m_sock, &buf[nDone], nChunk, MSG_NOSIGNAL);
#else
			min(nLen - nDone, m_nRecvChunk);
		int nRecv = recv(m_sock, &buf[nDone], nChunk, 0);
#endif

		if (nRecv <= 0)
		{
			if (nRecv == 0)
			{
				m_nError = -1;
			}
			else
			{
				UpdateError();
			}
			Close();
			return false;
		}
		nDone += nRecv;
	}
	return true;
}

bool QTCPSock::GenRecvTimed(void* pData, int nLen, int nTimeOut)
{
	int nDone = 0;
	char* buf =(char*) pData;
	if (nLen == 0) return true;

	while(nDone < nLen)
	{
#ifndef _WIN32
		pollfd fd;
		fd.fd = m_sock;
		fd.events = POLLIN | POLLERR;
		fd.revents = 0;
	
		int nRetVal = poll(&fd, 1, nTimeOut);
	
		if (nRetVal < 0)
		{
			m_nError = -2;
			Close();
			return false;
		}
		else if (nRetVal == 0)
		{
			m_nError = -3;
			Close();
			return false;
		}
#else
		fd_set readfds;
		timeval tv;
		FD_ZERO(&readfds);
		FD_SET(m_sock, &readfds);
		tv.tv_sec = (m_nTimeOut / 1000);
		tv.tv_usec = (m_nTimeOut % 1000) * 1000;
		int nRetVal = select(0, &readfds, NULL, NULL, &tv);

		if (nRetVal == SOCKET_ERROR)
		{
			m_nError = -2;
			Close();
			return false;
		}
		else if (nRetVal == 0)
		{
			m_nError = -3;
			Close();
			return false;
		}
#endif

		/* Attempt to prevent lockup  */
		int nAvail = 0;
		if (!GetInputSize(&nAvail))
		{
			return false;
		}

#ifndef _WIN32
		int nChunk = std::min(nLen - nDone, nAvail);
		int nRecv = recv(m_sock, &buf[nDone], nChunk, MSG_NOSIGNAL);
#else
		int nChunk = min(nLen - nDone, nAvail);
		int nRecv = recv(m_sock, &buf[nDone], nChunk, 0);
#endif
		if (nRecv <= 0)
		{
			if (nRecv == 0)
			{
				m_nError = -4;
			}
			else
			{
				UpdateError();
			}
			Close();
			return false;
		}
		nDone += nRecv;
	}
	return true;
}

bool QTCPSock::GenSend(struct msghdr& msgx, int nTotal)
{
	struct msghdr msg = msgx;
	while(true)
	{
		/* Send data */
		int nXRetVal = sendmsg(m_sock, &msg, MSG_NOSIGNAL | MSG_MORE);
		if (nXRetVal == -1)
		{
			/* Strange error, should be skipped */
			if (errno == EINTR)
			{
				continue;
			}

			//QLogError(0, "[DBG] Sending problem: LEN(%d), SENT(%d), SUM(%d), ERRNO(%d)", msg.msg_iovlen, nXRetVal, nTotal, errno);
			UpdateError();
			Close();
			return false;
		}

		/* Check that we have finished */
		if (nTotal == nXRetVal) return true;
		if (nTotal < nXRetVal)
		{
			QLogError(0, "[DBG] Bogus!");
			return true;
		}

		/* Update sizes */
		nTotal = nTotal - nXRetVal;
		while(nXRetVal > 0)
		{
			if ((int)msg.msg_iov->iov_len <= nXRetVal)
			{
				nXRetVal = nXRetVal - msg.msg_iov->iov_len;
				msg.msg_iov->iov_len = 0;
				msg.msg_iov++;
				msg.msg_iovlen--;
			}
			else
			{
				msg.msg_iov->iov_base = (void*)((uint8_t*)msg.msg_iov->iov_base + nXRetVal);
				msg.msg_iov->iov_len = msg.msg_iov->iov_len - nXRetVal;
				nXRetVal = 0;
			}
		}
	}
}

bool QTCPSock::GenSend(void* pData, int nLen)
{
	if (m_bTimed)
	{
		return GenSendTimed(pData, nLen, m_nTimeOut);
	}

	int nDone = 0;
	char* buf =(char*) pData;
	if (nLen == 0) return true;

	while(nDone < nLen)
	{
		/* Send */
#ifndef _WIN32
		int nChunk = std::min(nLen - nDone, m_nSendChunk);
		int nSent = send(m_sock, &buf[nDone], nChunk, MSG_NOSIGNAL);
#else
		int nChunk = min(nLen - nDone, m_nSendChunk);
		int nSent = send(m_sock, &buf[nDone], nChunk, 0);
#endif
		if (nSent <= 0)
		{
			if (nSent == 0)
			{
				m_nError = -5;
			}
			else
			{
				UpdateError();
			}
			Close();
			return false;
		}
		nDone += nSent;
	}
	return true;
}

bool QTCPSock::GenSendTimed(void* pData, int nLen, int nTimeOut)
{
	int nDone = 0;
	char* buf =(char*) pData;
	if (nLen == 0) return true;

	while(nDone < nLen)
	{
#ifndef _WIN32
		int nChunk = std::min(nLen - nDone, m_nSendChunk);
		pollfd fd;
		fd.fd = m_sock;
		fd.events = POLLOUT | POLLERR;
		fd.revents = 0;
	
		int nRetVal = poll(&fd, 1, nTimeOut);
	
		if (nRetVal < 0)
		{
			m_nError = -6;
			Close();
			return false;
		}
		else if (nRetVal == 0)
		{
			m_nError = -7;
			Close();
			return false;
		}

		int nSent = send(m_sock, &buf[nDone], nChunk, MSG_NOSIGNAL);
#else
		int nChunk = min(nLen - nDone, m_nSendChunk);
		fd_set writefds;
		timeval tv;
		FD_ZERO(&writefds);
		FD_SET(m_sock, &writefds);
		tv.tv_sec = (m_nTimeOut / 1000);
		tv.tv_usec = (m_nTimeOut % 1000) * 1000;
		int nRetVal = select(0, NULL, &writefds, NULL, &tv);

		if (nRetVal == SOCKET_ERROR)
		{
			m_nError = -5;
			Close();
			return false;
		}
		else if (nRetVal == 0)
		{
			m_nError = -6;
			Close();
			return false;
		}
		int nSent = send(m_sock, &buf[nDone], nChunk, 0);
#endif
		if (nSent <= 0)
		{
			if (nSent == 0)
			{
				m_nError = -7;
			}
			else
			{
				UpdateError();
			}
			Close();
			return false;
		}
		nDone += nSent;
	}
	return true;
}

bool QTCPSock::Open()
{
	// Set socket buffer size
#ifndef _WIN32
	if (m_sock >= 0)
#else
	if (m_sock != INVALID_SOCKET)
#endif
	{
		m_nError = 0;
		QLogWarning(0, "Trying to open already opened socket, ignored");
		return true;
	}

	m_sock = socket(AF_INET, SOCK_STREAM, 0);
#ifndef _WIN32
	if (m_sock == -1)
#else
	if (m_sock == INVALID_SOCKET)
#endif
	{
		UpdateError();
		QLogError(0, "Can not open socket: %d", m_nError);
		return false;
	}

	int opt = 1;
#ifndef _WIN32
	if (setsockopt(m_sock, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)))
#else
	if (setsockopt(m_sock, SOL_SOCKET, SO_REUSEADDR, (const char*)&opt, sizeof(opt)))
#endif
	{
		UpdateError();
		QLogError(0, "Can not set SO_REUSEADDR on socket: %d", m_nError);

		Close();
		return false;
	}
	linger lopt;
	lopt.l_onoff = 1;
	lopt.l_linger = 0;
#ifndef _WIN32
	if (setsockopt(m_sock, SOL_SOCKET, SO_LINGER, &lopt, sizeof(lopt)))
#else
	if (setsockopt(m_sock, SOL_SOCKET, SO_LINGER, (const char*)&opt, sizeof(opt)))
#endif
	{
		UpdateError();
		QLogError(0, "Can not set SO_LINGER on socket: %d", m_nError);

		Close();
		return false;
	}

	sockaddr_in addr;
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = 0;
	addr.sin_port = 0;
	if (bind(m_sock,(sockaddr *) &addr, sizeof(addr)))
	{
		UpdateError();
		QLogError(0, "Can not set bind socket: %d", m_nError);

		Close();
		return false;
	}

#ifndef _WIN32
	opt = 1;
	if (setsockopt(m_sock, IPPROTO_TCP, TCP_SYNCNT, (char*)&opt, sizeof(opt)))
	{
		UpdateError();
		QLogError(0, "Can not set TCP_SYNCNT on socket: %d", m_nError);

		Close();
		return false;
	}
#endif

	sockaddr_in saddr;
	saddr.sin_family = AF_INET;
	saddr.sin_addr.s_addr = m_bind.GetAddr();
	saddr.sin_port = htons(m_bind.GetPort());

	if (connect(m_sock,(sockaddr*)&saddr, sizeof(saddr)))
	{
		UpdateError();
		QLogWarning(0, "Can not connect to remote socket: %d", m_nError);

		Close();
		return false;
	}

	m_nError = 0;
	return true;
}

bool QTCPSock::GetInputSize(int* pAvail)
{
	if (!IsOpen()) return false;
#ifndef _WIN32
	if (ioctl(m_sock, FIONREAD, pAvail))
#else
	u_long nAvail = 0;
	if (ioctlsocket(m_sock, FIONREAD, &nAvail))
#endif
	{
		UpdateError();
		QLogError(0, "Can not FIONREAD on socket: %d", m_nError);

		Close();
		return false;
	}
#ifdef _WIN32
	else
	{
		*pAvail = nAvail;
	}
#endif
	return true;
}

bool QTCPSock::GetOutputSize(int* pAvail)
{
	if (!IsOpen()) return false;
#ifndef _WIN32
	if (ioctl(m_sock, TIOCOUTQ, pAvail, sizeof(*pAvail)))
#else
	/* Can not do anything */ // NLM // TODO
	*pAvail = 4096;
	if (0)
#endif
	{
		UpdateError();
		QLogError(0, "Can not TIOCOUTQ on socket: %d", m_nError);

		Close();
		return false;
	}
	return true;
}

void QTCPSock::SetNodelay()
{
	int nFlags = 1;
	if (setsockopt(m_sock, IPPROTO_TCP, TCP_NODELAY, (char*)&nFlags, sizeof(nFlags)) == -1)
	{
		UpdateError();
		QLogError(0, "Can not set TCP_NODELAY on socket: %d", m_nError);
		Close();
	}
}

void QTCPSock::SetCork(bool bSet)
{
#ifndef _WIN32
	int nFlags = bSet?1:0;
	if (setsockopt(m_sock, IPPROTO_TCP, TCP_CORK, (char*)&nFlags, sizeof(nFlags)) == -1)
	{
		UpdateError();
		QLogError(0, "Can not set TCP_CORK on socket: %d", m_nError);
		Close();
	}
#endif
}

bool QTCPSock::SetOTTBuffers()
{
	int nNewSize = 16 * 1024; /* 16 Kb buffer size */
	return SetBuffers(nNewSize, nNewSize);
}

bool QTCPSock::SetBuffers(int nSizeIn, int nSizeOut)
{
#ifndef _WIN32
	socklen_t len = sizeof(int);
#else
	int len = sizeof(int);
#endif

	if (nSizeIn > 0)
	{
#ifndef _WIN32
		if (setsockopt(m_sock, SOL_SOCKET, SO_RCVBUF, &nSizeIn, len) != 0)
#else
			if (setsockopt(m_sock, SOL_SOCKET, SO_RCVBUF, (const char*)&nSizeIn, len) != 0)
#endif
		{
			QLogError(0, "Error setting receive buffer %d", errno);
			Close();
			return false;
		}
	}

	if (nSizeOut > 0)
	{
#ifndef _WIN32
		if (setsockopt(m_sock, SOL_SOCKET, SO_SNDBUF, &nSizeOut, len) != 0)
#else
			if (setsockopt(m_sock, SOL_SOCKET, SO_SNDBUF, (const char*)&nSizeOut, len) != 0)
#endif
		{
			QLogError(0, "Error setting send buffer %d", errno);
			Close();
			return false;
		}
	}

	return true;
}

QNetAddr* QTCPSock::GetBounds()
{
	return &m_bind;
}

bool QTCPSock::GetPeer(QNetAddr* pAddr)
{
	return pAddr->SetAddr(m_sock, true);
}

int QTCPSock::GetTimeOut()
{
	return m_nTimeOut;
}

int QTCPSock::GetError()
{
	return m_nError;
}

#ifndef _WIN32
int QTCPSock::GetSock()
#else
SOCKET QTCPSock::GetSock()
#endif
{
#ifndef _WIN32
	return __sync_add_and_fetch(&m_sock, 0);
#else
	return m_sock;
#endif
}

#ifndef _WIN32
int QTCPSock::SetSock(int nSock)
#else
SOCKET QTCPSock::SetSock(SOCKET nSock)
#endif
{
#ifndef _WIN32
	int nRetVal = __sync_lock_test_and_set(&m_sock, nSock);
#else
	SOCKET nRetVal = m_sock;
	m_sock = nSock;
#endif
	m_bind.SetAddr(m_sock);
	return nRetVal;
}

void QTCPSock::SetTimed(bool bTimed)
{
	m_bTimed = bTimed;
}

void QTCPSock::SetTimeOut(int nTimeout)
{
	m_nTimeOut = nTimeout;
}

void QTCPSock::SetHWKA()
{
	int nKeepAlive = 1;
	int nKeepCount = 5;
	int nKeepIdle = 30;
	int nKeepInt = 5;

	if (!IsOpen())
	{
		return;
	}

#ifndef _WIN32
	if (setsockopt(m_sock, SOL_SOCKET, SO_KEEPALIVE, &nKeepAlive, sizeof (nKeepAlive)))
#else
	if (setsockopt(m_sock, SOL_SOCKET, SO_KEEPALIVE, (const char*)&nKeepAlive, sizeof (nKeepAlive)))
#endif
	{
		UpdateError();
		QLogError(0, "Could not set SO_KEEPALIVE on incoming socket: %d", m_nError);
		Close();
		return;
	}

#ifndef _WIN32
	if (setsockopt(m_sock, IPPROTO_TCP, TCP_KEEPCNT, &nKeepCount, sizeof (nKeepCount)))
	{
		UpdateError();
		QLogError(0, "Could not set TCP_KEEPCNT on incoming socket: %d", errno);
		Close();
		return;
	}

	if (setsockopt(m_sock, IPPROTO_TCP, TCP_KEEPIDLE, &nKeepIdle, sizeof (nKeepIdle)))
	{
		UpdateError();
		QLogError(0, "Could not set TCP_KEEPIDLE on incoming socket: %d", errno);
		Close();
		return;
	}

	if (setsockopt(m_sock, IPPROTO_TCP, TCP_KEEPINTVL, &nKeepInt, sizeof (nKeepInt)))
	{
		UpdateError();
		QLogError(0, "Could not set TCP_KEEPINTVL on incoming socket: %d", errno);
		Close();
		return;
	}
#else
	tcp_keepalive ka;
	ka.onoff = 1;
	ka.keepalivetime = 5000;
	ka.keepaliveinterval = 300;
#endif
}

bool QTCPSock::CheckConnect()
{
	if (!IsOpen()) return false;

#ifndef _WIN32
	// Now do POLL
	pollfd fd;
	fd.fd = m_sock;
	fd.events = POLLERR | POLLRDHUP | POLLHUP | POLLNVAL;
	fd.revents = 0;

	int nRetVal = poll(&fd, 1, 0);

	if (nRetVal < 0)
	{
		UpdateError();
		Close();
		return false;
	}
	else if (nRetVal > 0)
	{
		m_nError = 0;
		Close();
		return false;
	}
#else
	fd_set checkfds;
	timeval tv;
	FD_ZERO(&checkfds);
	FD_SET(m_sock, &checkfds);
	tv.tv_sec = 0;
	tv.tv_usec = 0;
	int nRetVal = select(0, NULL, NULL, &checkfds, &tv);

	if (nRetVal == SOCKET_ERROR)
	{
		m_nError = -5;
		Close();
		return false;
	}
	else if (nRetVal == 1)
	{
		Close();
		return false;
	}
#endif
	return true;
}

bool QTCPSock::GetTimed()
{
	return m_bTimed;
}

void QTCPSock::UpdateError()
{
#ifndef _WIN32
	m_nError = errno;
#else
	m_nError = ::WSAGetLastError();
#endif
}
